package com.sunlabs.archsim.fleet.ships.fifo;

import java.util.Iterator;
import java.util.LinkedList;

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.engine.Director;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.TriggerAction;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

import junit.framework.TestCase;
import junit.textui.TestRunner;

/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/** 
 * The FIFO run by JUnit framework.
 * 
 * @author ib27688
 *
 */
public class FifoTest extends TestCase {

	public static void main(String[] args) {
		TestRunner.run(FifoTest.class);
	} // main
	
	public void setUp() {
	}
	
	public void tearDown() {
		Director.getInstance().masterClear();
	}
	
	public FifoTest(String arg) {
		super(arg);
	}
	
	// vanilla test for Fifo of capacity 5
	public void testFifoShip_5_10_10() {
		testFifoShip(20, 5, 10, 10);
	}
	
	// vanilla test for fifo of capacity 1
	public void testFifoShip_1_10_10() {
		testFifoShip(20, 1, 10, 10);
	}
	
	// vanilla test for fifo of capacity 0
	public void testFifoShip_0_10_10() {
		testFifoShip(20, 0, 10, 10);
	}
	
	// aggressive source
	public void testFifoShip_5_0_10() {
		testFifoShip(40, 5, 0, 10);
	}		

	// aggressive source, capacity 1
	public void testFifoShip_1_0_10() {
		testFifoShip(20, 1, 0, 10);
	}		

	// aggressive source, capacity 0
	public void testFifoShip_0_0_10() {
		testFifoShip(20, 0, 0, 10);
	}		
	
	// aggressive sink, capacity 5
	public void testFifoShip_5_10_0() {
		testFifoShip(40, 5, 10, 0);
	}		

	// aggressive sink, capacity 1
	public void testFifoShip_1_10_0() {
		testFifoShip(20, 1, 10, 0);
	}		

	// aggressive sink, capacity 0
	public void testFifoShip_0_10_0() {
		testFifoShip(20, 0, 10, 0);
	}		


	
	
	
	// vanilla test for Fifo of capacity 5
	public void testFifo_5_10_10() {
		testFifo(20, 5, 10, 10);
	}
	
	// vanilla test for fifo of capacity 1
	public void testFifo_1_10_10() {
		testFifo(20, 1, 10, 10);
	}
	
	// vanilla test for fifo of capacity 0
	public void testFifo_0_10_10() {
		testFifo(20, 0, 10, 10);
	}
	
	// aggressive source
	public void testFifo_5_0_10() {
		testFifo(40, 5, 0, 10);
	}		

	// aggressive source, capacity 1
	public void testFifo_1_0_10() {
		testFifo(20, 1, 0, 10);
	}		

	// aggressive source, capacity 0
	public void testFifo_0_0_10() {
		testFifo(20, 0, 0, 10);
	}		
	
	// aggressive sink, capacity 5
	public void testFifo_5_10_0() {
		testFifo(40, 5, 10, 0);
	}		

	// aggressive sink, capacity 1
	public void testFifo_1_10_0() {
		testFifo(20, 1, 10, 0);
	}		

	// aggressive sink, capacity 0
	public void testFifo_0_10_0() {
		testFifo(20, 0, 10, 0);
	}		
	
	
	
	/**
	 * Test driver
	 * @param n number if items to send
	 * @param capacity fifo capacity
	 * @param srcDelay delay for the source
	 * @param sinkDelay delay for the sink
	 */	
	public void testFifoShip(int n, int capacity, int srcDelay, int sinkDelay) {
		FifoShipTestGroup test= new FifoShipTestGroup(n,capacity, srcDelay, sinkDelay);
		Director director= Director.getInstance();
		// run selfTest
		Boolean selfCheck= test.selfCheck();
		assertTrue(selfCheck);
		
		// run the rest of the test only if self-check passed
		if (selfCheck) {
			director.reset();
			// run for long enough to go through all the tests
			director.run();
			boolean testsPassed= test.testCheck();
			assertTrue(testsPassed);
		}
		
	}

	/**
	 * Test driver
	 * @param n number if items to send
	 * @param capacity fifo capacity
	 * @param srcDelay delay for the source
	 * @param sinkDelay delay for the sink
	 */	
	public void testFifo(int n, int capacity, int srcDelay, int sinkDelay) {
		FifoTestGroup test= new FifoTestGroup(n,capacity, srcDelay, sinkDelay);
		Director director= Director.getInstance();
		// run selfTest
		Boolean selfCheck= test.selfCheck();
		assertTrue(selfCheck);
		
		// run the rest of the test only if self-check passed
		if (selfCheck) {
			director.reset();
			// run for long enough to go through all the tests
			director.run();
			boolean testsPassed= test.testCheck();
			assertTrue(testsPassed);
		}
		
	}
	
} // FifoTest

/**
 * Test setup for a fifo ship test
 * 
 * @param n - the number of data items the source will produce
 * @param capacity - the capacity of the fifo
 * 
 * @author ib27688
 */
class FifoShipTestGroup extends CompositeEntity {
	
	int count;
	
	FifoSource src;
	FifoShipSink sink;
	FifoShip fifo;
	
	public FifoShipTestGroup(int n, int capacity, int srcDelay, int sinkDelay) {
		super("FifoShipTestGroup" + n);
		
		count= n;
		
		src= new FifoSource("FifoShipTestSource_", n, srcDelay);
		addMember(src);
		
		fifo= new FifoShip("TestFifoShip", capacity);
		addMember(fifo);
		
		sink= new FifoShipSink("FifoShipTestSink_", n, sinkDelay);
		addMember(sink);
		
		// states that connect thet fifo with the source and sink
		State srcDataState= new State("FifoShipTestDataInState_" + capacity);
		State srcAckDataState= new State("FifoShipTestAckDataInState_" + capacity);
		
		srcDataState.attach(src.getTerminal(FleetGlobal.OUT));
		srcDataState.attach(fifo.getTerminal(FleetGlobal.IN));
		
		srcAckDataState.attach(src.getTerminal(FleetGlobal.ACK_OUT));
		srcAckDataState.attach(fifo.getTerminal(FleetGlobal.ACK_IN));
		
		State sinkDataState= new State("FifoShipTestDataOutState_" + capacity);
		State sinkAckDataState= new State("FifoShipTestDataOutState_" + capacity);
		
		sinkDataState.attach(fifo.getTerminal(FleetGlobal.OUT));
		sinkDataState.attach(sink.getTerminal(FleetGlobal.IN));
		
		sinkAckDataState.attach(fifo.getTerminal(FleetGlobal.ACK_OUT));
		sinkAckDataState.attach(sink.getTerminal(FleetGlobal.ACK_IN));
		
		srcAckDataState.setInitiallyActive(new Data());
		sinkAckDataState.setInitiallyActive(new Data());
	} //FifoTestGroup constructor
	
	boolean testCheck() {
		LinkedList<Data> computedResults= new LinkedList<Data>();
		for (int i= 0; i<count; i++) {
			computedResults.addLast(new Data(i));
		}
		
		
		/* display computed results */
		/*
		System.out.println("\nFifo test computed results: ");
		Iterator<Data> ci= computedResults.iterator();
		while (ci.hasNext()) {
			Data d= ci.next();
			System.out.println(d);
		}
		*/
		LinkedList<Data> simulatedResults= sink.getResults();
		
		
		/* display simulated results */
		/*
		System.out.println("\n\nFifo test simulated results: ");
		Iterator<Data> si= simulatedResults.iterator();
		while (si.hasNext()) {
			Data d= si.next();
			System.out.println(d);
		}
		*/
		return computedResults.equals(simulatedResults);
	} // testCheck
	
} // FifoTestGroup


/**
 * Test setup for a fifo test
 * 
 * @param n - the number of data items the source will produce
 * @param capacity - the capacity of the fifo
 * 
 * @author ib27688
 */
class FifoTestGroup extends CompositeEntity {
	
	int count;
	
	FifoSource src;
	FifoSink sink;
	Fifo fifo;
	
	public FifoTestGroup(int n, int capacity, int srcDelay, int sinkDelay) {
		super("FifoShipTestGroup" + n);
		
		count= n;
		
		src= new FifoSource("FifoShipTestSource_", n, srcDelay);
		addMember(src);
		
		fifo= new Fifo("TestFifoShip", capacity);
		addMember(fifo);
		
		sink= new FifoSink("FifoShipTestSink_", n, sinkDelay);
		addMember(sink);
		
		// states that connect thet fifo with the source and sink
		State srcDataState= new State("FifoTestDataInState_" + capacity);
		State srcAckDataState= new State("FifoTestAckDataInState_" + capacity);
		
		srcDataState.attach(src.getTerminal(FleetGlobal.OUT));
		srcDataState.attach(fifo.getTerminal(FleetGlobal.IN));
		
		srcAckDataState.attach(src.getTerminal(FleetGlobal.ACK_OUT));
		srcAckDataState.attach(fifo.getTerminal(FleetGlobal.ACK_IN));
		
		State sinkDataState= new State("FifoTestDataOutState_" + capacity);
		State sinkAckDataState= new State("FifoTestDataOutState_" + capacity);
		
		sinkDataState.attach(fifo.getTerminal(FleetGlobal.OUT));
		sinkDataState.attach(sink.getTerminal(FleetGlobal.IN));
		
		sinkAckDataState.attach(fifo.getTerminal(FleetGlobal.ACK_OUT));
		sinkAckDataState.attach(sink.getTerminal(FleetGlobal.ACK_IN));
		
		srcAckDataState.setInitiallyActive(new Data());
		sinkAckDataState.setInitiallyActive(new Data());
	} //FifoTestGroup constructor
	
	boolean testCheck() {

		LinkedList<Payload> computedResults= new LinkedList<Payload>();
		for (int i= 0; i<count; i++) {
			Payload p= new Payload();
			p.data= new Data(i);
			computedResults.addLast(p);
		}
		
		/* display computed results */
		/*
		System.out.println("\nFifo test computed results: ");
		Iterator<Data> ci= computedResults.iterator();
		while (ci.hasNext()) {
			Data d= ci.next();
			System.out.println(d);
		}
		*/
		LinkedList<Payload> simulatedResults= sink.getResults();
		
		
		/* display simulated results */
		/*
		System.out.println("\n\nFifo test simulated results: ");
		Iterator<Data> si= simulatedResults.iterator();
		while (si.hasNext()) {
			Data d= si.next();
			System.out.println(d);
		}
		*/
		return computedResults.equals(simulatedResults);
	} // testCheck
	
} // FifoTestGroup



/**
 * Data source for fifo under test
 * @author ib27688
 *
 */
class FifoSource extends Component {
	
	Terminal out;
	Terminal ackOut;
	
	FifoSourceAction srcAction;
	
	public FifoSource(String name, int n, int srcDelay) {
		super(name, null);
		
		srcAction= new FifoSourceAction("FifoSourceAction", n, srcDelay);
		addMember(srcAction);
		out= new OutputTerminal(FleetGlobal.OUT, this);
		ackOut= new InputTerminal(FleetGlobal.ACK_OUT, this);
		//attach the terminals
		srcAction.attach(out);
		srcAction.attach(ackOut);
	} // FifoSource
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.OUT)) {
			return out;
		}
		else if (name.equals(FleetGlobal.ACK_OUT)) {
			return ackOut;
		}
		else {
			// unknown terminal
			return null;
		}
	} // getTerminal
	
} // FifoSource


class FifoSourceAction extends TriggerAction {

	int count;
	public FifoSourceAction(String name, int n, int delay) {
		super(name);
		// how many data items to produce
		count= n;
		enabled= true;
		outputFunctions.put(FleetGlobal.OUT, new GenerateData(delay));

		// enableTrigggerCommand.execute(null);
		
	} // FifoSourceAction
	
	/** Generate new data items */
	private class GenerateData extends Function {

		private int next= 0;
		
		public GenerateData(int delay) {
			super("GenerateData");
			functionDelay= new Delay(delay); // was 20
		}
		
		public Object value() {
			int current= next;
			next++;
			if (next >= count) {
				// done counting, disable the source
				enabled= false;
			}
			if (FifoGlobal.sourceDisplayGenerated) {
				director.journal("Source generated " + next);
			}
			
			Payload p= new Payload();
			Data d= new Data(current);
			p.data= d;			
			return p;	
		} // value
	} // PassData

} // FifoSourceAction

/**
 * Sink for data for a Fifo under test
 * @author ib27688
 *
 */
class FifoShipSink extends Component {
	Terminal in;
	Terminal ackIn;
	
	FifoShipSinkAction sinkAction;
	
	public FifoShipSink(String name, int n, int sinkDelay) {
		super(name, null);
		
		sinkAction= new FifoShipSinkAction("FifoSinkAction", n, sinkDelay);
		addMember(sinkAction);
		in= new InputTerminal(FleetGlobal.IN, this);
		ackIn= new OutputTerminal(FleetGlobal.ACK_IN, this);
		//attach the terminals
		sinkAction.attach(in);
		sinkAction.attach(ackIn);
	} // FifoSource
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.IN)) {
			return in;
		}
		else if (name.equals(FleetGlobal.ACK_IN)) {
			return ackIn;
		}
		else {
			// unknown terminal
			return null;
		}
	} // getTerminal
	
	/**
	 * Get results...
	 * @return a list of results of tests
	 */
	LinkedList<Data> getResults() {
		return sinkAction.results;
	} // getResults
		
	
} // class FifoShipSink



class FifoShipSinkAction extends Action {
	
	// the number or items to receive
	int count;
	
	LinkedList<Data> results;
	
	
	public FifoShipSinkAction(String name, int n, int delay) {
		super(name);
		// how many data items to receive
		count= n;
		// list of received results
		results= new LinkedList<Data>();
		outputFunctions.put(FleetGlobal.ACK_IN, new ActionFired(delay));
	} // FifoSinkAction
	
	/** Vacuous function used to acknowldge that the action has fired */ 
	private class ActionFired extends OutputFunction {
		
		public ActionFired(int d) {
			super("Vacous");
			functionDelay= new Delay(d); 
		}
	
		public Data value() {
			Iterator ti= terminals.iterator();
			Terminal t= null;
			boolean done= false;
			while (!done && ti.hasNext()) {
				t= (Terminal)(ti.next());
				if (t.getBaseName() == FleetGlobal.IN) {
					// found the terminal
					done= true;
					Data datum= (Data)t.getData();
					results.addLast(datum);
				}
			}
			if (t != null && FifoGlobal.sinkDisplayReceived ) {
				director.journal("Sink received " + ((Data)t.getData()).value);
			}
			return null;
		} // value
	} // AcknowledgeData

} // FifoShipSinkAction




/**
 * Sink for data for a Fifo under test
 * @author ib27688
 *
 */
class FifoSink extends Component {
	Terminal in;
	Terminal ackIn;
	
	FifoSinkAction sinkAction;
	
	public FifoSink(String name, int n, int sinkDelay) {
		super(name, null);
		
		sinkAction= new FifoSinkAction("FifoSinkAction", n, sinkDelay);
		addMember(sinkAction);
		in= new InputTerminal(FleetGlobal.IN, this);
		ackIn= new OutputTerminal(FleetGlobal.ACK_IN, this);
		//attach the terminals
		sinkAction.attach(in);
		sinkAction.attach(ackIn);
	} // FifoSource
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.IN)) {
			return in;
		}
		else if (name.equals(FleetGlobal.ACK_IN)) {
			return ackIn;
		}
		else {
			// unknown terminal
			return null;
		}
	} // getTerminal
	
	/**
	 * Get results...
	 * @return a list of results of tests
	 */
	LinkedList<Payload> getResults() {
		return sinkAction.results;
	} // getResults
		
	
} // class FifoSink




class FifoSinkAction extends Action {
		
		// the number or items to receive
		int count;
		
		LinkedList<Payload> results;
		
		
		public FifoSinkAction(String name, int n, int delay) {
			super(name);
			// how many data items to receive
			count= n;
			// list of received results
			results= new LinkedList<Payload>();
			outputFunctions.put(FleetGlobal.ACK_IN, new ActionFired(delay));
		} // FifoSinkAction
		
		/** Vacuous function used to acknowldge that the action has fired */ 
		private class ActionFired extends OutputFunction {
			
			public ActionFired(int d) {
				super("Vacous");
				functionDelay= new Delay(d); 
			}
		
			public Data value() {
				Iterator ti= terminals.iterator();
				Terminal t= null;
				boolean done= false;
				while (!done && ti.hasNext()) {
					t= (Terminal)(ti.next());
					if (t.getBaseName() == FleetGlobal.IN) {
						// found the terminal
						done= true;
						Payload p= (Payload)t.getData();
						results.addLast(p);
					}
				}
				if (t != null && FifoGlobal.sinkDisplayReceived ) {
					director.journal("Sink received " + ((Payload)t.getData()));
				}
				return null;
			} // value
		} // AcknowledgeData
	
} // FifoSinkAction